import os, sys as System, time
import sys

from rfc822 import Message as Msg822
from mimetools import Message as MimeMsg, decode, choose_boundary as MimeBoundary

import test.test_StringIO as StringTest

for k, v in globals().items():
    print `k`, v
